import os
import timeit
import logging
import subprocess
import re

import gmsh

from fiqus.utils.Utils import FilesAndFolders
from fiqus.utils.Utils import GmshUtils
from fiqus.data.RegionsModelFiQuS import RegionsModel
from fiqus.pro_assemblers.ProAssembler import ASS_PRO as ass_pro
from fiqus.mains.MainPancake3D import Base
from fiqus.data.DataFiQuSPancake3D import Pancake3DGeometry, Pancake3DMesh
from fiqus.parsers.ParserRES import ParserRES

logger = logging.getLogger(__name__)


class Solve(Base):
    """
    Main class to run GetDP for Pancake3D.

    :param fdm: FiQuS data model
    :param GetDP_path: Settings for GetDP
    :type GetDP_path: dict
    """

    def __init__(
        self,
        fdm,
        GetDP_path: str = None,
        geom_folder=None,
        mesh_folder=None,
        solution_folder=None,
    ) -> None:
        super().__init__(fdm, geom_folder, mesh_folder, solution_folder)

        self.GetDP_path = GetDP_path  # Some settings for GetDP
        # check for init from res file option
        if self.dm.magnet.solve.initFromPrevious:
            self.res_file_path = os.path.join(
                self.mesh_folder,
                f"Solution_{self.dm.magnet.solve.initFromPrevious}",
                self.magnet_name + ".res",
            )

            if not os.path.isfile(self.res_file_path):
                raise ValueError(f"Res file {self.res_file_path} does not exist.")
            else:
                if self.dm.run.type not in ["solve_only", "solve_with_post_process_python"]:
                    raise ValueError(
                        f"Run type should be solve only for init from res file option."
                    )

                if self.dm.magnet.solve.type == "weaklyCoupled":
                    _no_of_previous_solutions = 2
                else:
                    _no_of_previous_solutions = 1

                logger.info(
                    f"Initializing from previous solution {self.res_file_path}."
                )

                # parse given res file
                parsed_init_res = ParserRES(self.res_file_path)

                # remove all but no_of_previous_solutions
                parsed_init_res.solution["time_real"] = parsed_init_res.solution[
                    "time_real"
                ][-_no_of_previous_solutions:]
                parsed_init_res.solution["time_imag"] = parsed_init_res.solution[
                    "time_imag"
                ][-_no_of_previous_solutions:]
                parsed_init_res.solution["time_step"] = parsed_init_res.solution[
                    "time_step"
                ][-_no_of_previous_solutions:]
                parsed_init_res.solution["dof_data"] = parsed_init_res.solution[
                    "dof_data"
                ][-_no_of_previous_solutions:]
                parsed_init_res.solution["solution"] = parsed_init_res.solution[
                    "solution"
                ][-_no_of_previous_solutions:]

                self.res_file_without_previous_solutions = os.path.join(
                    self.solution_folder, self.magnet_name + ".res"
                )
                ParserRES(self.res_file_without_previous_solutions, parsed_init_res)

        # Create pro file:
        self.ap = ass_pro(os.path.join(self.solution_folder, self.magnet_name))

        # Start GMSH:
        self.gu = GmshUtils(self.mesh_folder)
        self.gu.initialize()

        # Read regions model:
        self.rm = FilesAndFolders.read_data_from_yaml(self.regions_file, RegionsModel)

        # Check if the geometry data and mesh data has not been altered after they are
        # created:
        previousGeo = FilesAndFolders.read_data_from_yaml(
            self.geometry_data_file, Pancake3DGeometry
        )
        previousMesh = FilesAndFolders.read_data_from_yaml(
            self.mesh_data_file, Pancake3DMesh
        )

        if previousGeo.dict() != self.geo.dict():
            raise ValueError(
                "Geometry data has been changed. Please regenerate the geometry or load"
                " the previous geometry data."
            )
        elif previousMesh.dict() != self.mesh.dict():
            raise ValueError(
                "Mesh data has been changed. Please regenerate the mesh or load the"
                " previous mesh data."
            )

    def assemble_pro(self):
        logger.info(f"Assembling pro file ({self.pro_file}) has been started.")
        start_time = timeit.default_timer()

        self.ap.assemble_combined_pro(
            template=self.solve.proTemplate,
            rm=self.rm,
            dm=self.dm,
        )

        logger.info(
            f"Assembling pro file ({self.pro_file}) has been finished in"
            f" {timeit.default_timer() - start_time:.2f} s."
        )

    def run_getdp(self, solve=True, postOperation=True):
        logger.info("Solving Pancake3D magnet has been started.")
        start_time = timeit.default_timer()

        getdpArguments = ["-v2", "-mat_mumps_cntl_1", "0", "-verbose", "6"]

        if self.dm.magnet.solve.initFromPrevious:
            getdpArguments.extend(
                ["-restart", "-res", str(self.res_file_without_previous_solutions)]
            )

        # Add solve argument
        if solve:
            getdpArguments.extend(["-solve", f"RESOLUTION_{self.solve.type}"])

        # Add post operation argument
        if postOperation:
            posStringList = ["-pos"]
            if solve is False:
                # Quantities to be saved:
                if self.dm.magnet.solve.save is not None:
                    for quantity in self.dm.magnet.solve.save:
                        posStringList.append(f"{quantity.getdpPostOperationName}")

                if self.dm.magnet.postproc is not None:
                    if self.dm.magnet.postproc.timeSeriesPlots is not None:
                        # Post-operations for Python post-processing:
                        for timeSeriesPlot in self.dm.magnet.postproc.timeSeriesPlots:
                            posStringList.append(
                                f"POSTOP_timeSeriesPlot_{timeSeriesPlot.quantity}"
                            )

                    if self.dm.magnet.postproc.magneticFieldOnCutPlane is not None:
                        # Post-operations for Python post-processing:
                        posStringList.append("POSTOP_magneticFieldOnCutPlaneVector")
                        posStringList.append("POSTOP_magneticFieldOnCutPlaneMagnitude")

            else:
                posStringList.append("POSTOP_dummy")
            getdpArguments.extend(posStringList)

        # Add mesh argument
        getdpArguments.extend(["-msh", f"{self.mesh_file}"])

        # Add pre-processing argument
        getdpArguments.extend(["-pre", f"RESOLUTION_{self.solve.type}"])

        getdp_binary = self.GetDP_path

        getdpProcess = subprocess.Popen(
            [getdp_binary, self.pro_file] + getdpArguments,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        with getdpProcess.stdout:
            for line in iter(getdpProcess.stdout.readline, b""):
                line = line.decode("utf-8").rstrip()
                line = line.split("\r")[-1]
                if "Info" in line:
                    parsedLine = re.sub(r"Info\s*:\s*", "", line)
                    logger.info(parsedLine)
                elif "Warning" in line:
                    parsedLine = re.sub(r"Warning\s*:\s*", "", line)
                    if "Unknown" not in parsedLine:
                        logger.warning(parsedLine)
                elif "Error" in line:
                    parsedLine = re.sub(r"Error\s*:\s*", "", line)
                    logger.error(parsedLine)
                    logger.error("Solving Pancake3D magnet has failed.")
                    raise Exception(parsedLine)
                elif "Critical" in line:
                    parsedLine = re.sub(r"Critical\s*:\s*", "", line)
                    if "Quench started!" in parsedLine:
                        logger.critical(r" _____ _  ___        ____ ")
                        logger.critical(r"|  ___(_)/ _ \ _   _/ ___| ")
                        logger.critical(r"| |_  | | | | | | | \___ \ ")
                        logger.critical(r"|  _| | | |_| | |_| |___) |")
                        logger.critical(r"|_|   |_|\__\_\\__,_|____/ ")
                        logger.critical("")
                        logger.critical("The coil has been quenched!")
                    else:
                        logger.critical(parsedLine)
                # else:
                #     logger.info(line)

        getdpProcess.wait()

        logger.info(
            "Solving Pancake3D magnet has been finished in"
            f" {timeit.default_timer() - start_time:.2f} s."
        )

        if self.solve_gui:
            gmsh.option.setNumber("Geometry.Volumes", 0)
            gmsh.option.setNumber("Geometry.Surfaces", 0)
            gmsh.option.setNumber("Geometry.Curves", 0)
            gmsh.option.setNumber("Geometry.Points", 0)
            posFiles = [
                fileName
                for fileName in os.listdir(self.solution_folder)
                if fileName.endswith(".pos")
            ]
            for posFile in posFiles:
                gmsh.open(os.path.join(self.solution_folder, posFile))
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()



import functools
import itertools
import operator
import os
import unittest
from pathlib import Path

import numpy as np

from fiqus.data.DataFiQuS import FDM
from fiqus.utils.Utils import FilesAndFolders as FFs


class Paths:
    """
    Helper class used in FiQuS tests to get file and folder paths
    :param model_name: name of yaml input file (without .yaml)
    :param f_extension: file extension to apply to the folders, for example '.brep' or '.msh'
    :return: tuple: fdm, outputs_folder, input_folder, input_file, model_folder, model_file, reference_folder, reference_file
    """

    def __init__(self, model_name, f_extension=''):
        self.inputs_folder_name = '_inputs'
        self.outputs_folder_name = '_outputs'
        self.references_folder_name = '_references'

        self.test_outputs_folder = os.path.join(os.getcwd(), self.outputs_folder_name)
        self.inputs_folder = os.path.join(os.getcwd(), self.inputs_folder_name, model_name)
        self.model_folder = os.path.join(os.getcwd(), self.outputs_folder_name, model_name)
        self.references_folder = os.path.join(os.getcwd(), self.references_folder_name, model_name)

        self.input_file = os.path.join(self.inputs_folder, f'{model_name}.yaml')
        self.model_file = os.path.join(self.model_folder, f'{model_name}.{f_extension}')
        self.reference_file = os.path.join(self.references_folder, f'{model_name}.{f_extension}')


def filecmp(filename1, filename2):
    """
    From: https://stackoverflow.com/questions/254350/in-python-is-there-a-concise-way-of-comparing-whether-the-contents-of-two-text
    Do the two files have exactly the same contents?
    """
    with open(filename1, "rb") as fp1, open(filename2, "rb") as fp2:
        print(f'The {filename1} size is: {os.fstat(fp1.fileno()).st_size} b')
        print(f'The {filename2} size is: {os.fstat(fp2.fileno()).st_size} b')
        if os.fstat(fp1.fileno()).st_size != os.fstat(fp2.fileno()).st_size:
            return False  # different sizes ∴ not equal

        # set up one 4k-reader for each file
        fp1_reader = functools.partial(fp1.read, 4096)
        fp2_reader = functools.partial(fp2.read, 4096)

        # pair each 4k-chunk from the two readers while they do not return '' (EOF)
        cmp_pairs = zip(iter(fp1_reader, b''), iter(fp2_reader, b''))

        # return True for all pairs that are not equal
        inequalities = itertools.starmap(operator.ne, cmp_pairs)
        ineqs = []
        for ineq in inequalities:
            ineqs.append(ineq)
        # voila; any() stops at first True value
        print(f'The file comp function gives: {not any(inequalities)}')
        return not any(inequalities)


def assert_two_parameters(true_value, test_value):
    """
     Some functions used in multiple test functions
        **Assert two parameters - accepts multiple types**
    """
    # TODO: improve robustness and readability
    test_case = unittest.TestCase()

    if isinstance(true_value, np.ndarray) or isinstance(true_value, list):
        if len(true_value) == 1:
            true_value = float(true_value)

    if isinstance(test_value, np.ndarray) or isinstance(test_value, list):
        if len(test_value) == 1:
            test_value = float(test_value)

    # Comparison
    if isinstance(test_value, np.ndarray) or isinstance(test_value, list):
        if np.array(true_value).ndim == 2:
            for i, test_row in enumerate(test_value):
                if isinstance(test_row[0], np.floating):
                    test_row = np.array(test_row).round(10)
                    true_value[i] = np.array(true_value[i]).round(10)

                test_case.assertListEqual(list(test_row), list(true_value[i]))
        else:
            if isinstance(test_value[0], np.floating):
                test_value = np.array(test_value).round(10)
                true_value = np.array(true_value).round(10)

            test_case.assertListEqual(list(test_value), list(true_value))
    else:
        test_case.assertEqual(test_value, true_value)



import unittest
from tests.utils.fiqus_test_classes import FiQuSSolverTests
from fiqus.data.DataFiQuS import FDM


class TestSolvers(FiQuSSolverTests):
    def test_Pancake3D(self):
        """
        Checks if Pancake3D solvers work correctly by comparing the results to the
        reference results that were checked manually.
        """
        with self.subTest(msg="tsa linear weakly coupled"):
            model_name = "TEST_Pancake3D_TSA"
            data_model: FDM = self.get_data_model(model_name)

            data_model.magnet.solve.type = "weaklyCoupled"

            self.solve(data_model, model_name)

            # Compare the pro files:
            pro_file = self.get_path_to_generated_file(
                file_name=model_name, file_extension="pro"
            )
            reference_pro_file = self.get_path_to_reference_file(
                file_name=model_name, file_extension="pro"
            )
            self.compare_text_files(pro_file, reference_pro_file)

            # Compare the results files:
            pos_file = self.get_path_to_generated_file(file_name="Temperature-DefaultFormat", file_extension="pos")
            reference_pos_file = self.get_path_to_reference_file(file_name="Temperature-DefaultFormat", file_extension="pos")
            self.compare_pos_files(pos_file, reference_pos_file)

        with self.subTest(msg="tsa linear strongly coupled"):
            model_name = "TEST_Pancake3D_TSA"
            data_model = self.get_data_model(model_name)

            data_model.magnet.solve.type = "stronglyCoupled"
            data_model.magnet.solve.ti.cooling = "fixedTemperature"

            self.solve(data_model, model_name)

        with self.subTest(msg="tsa linear electromagnetic"):
            model_name = "TEST_Pancake3D_TSA"
            data_model = self.get_data_model(model_name)

            data_model.magnet.solve.type = "electromagnetic"
            del data_model.magnet.solve.t.adaptive.postOperationTolerances[1]

            self.solve(data_model, model_name)


if __name__ == "__main__":
    unittest.main()



import unittest
from tests.utils.fiqus_test_classes import FiQuSMeshTests

class TestMeshGenerators(FiQuSMeshTests):
    def test_Pancake3D(self):
        """
        Checks if Pancake3D geometry generators work correctly by comparing the number
        of entities in the generated geometry file to the reference file that was
        checked manually.
        """
        model_names = [
            "TEST_Pancake3D_REF",
            "TEST_Pancake3D_REFStructured",
            "TEST_Pancake3D_TSA",
            "TEST_Pancake3D_TSAStructured",
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                # data_model can be modified here if necessary
                # Example:

                # data_model.magnet.mesh.wi.axne = 30

                self.generate_mesh(data_model, model_name)

                # Compare the number of entities with the reference file:
                mesh_file = self.get_path_to_generated_file(file_name=model_name, file_extension="msh")
                reference_file = self.get_path_to_reference_file(model_name, "msh")
                self.compare_mesh_qualities(mesh_file, reference_file)

                # Compare the regions files:
                regions_file = self.get_path_to_generated_file(file_name=model_name, file_extension="regions")
                reference_regions_file = self.get_path_to_reference_file(
                    model_name, "regions"
                )
                self.compare_json_or_yaml_files(regions_file, reference_regions_file)


if __name__ == "__main__":
    unittest.main()



import re
import math


class ParserGetDPTimeTable:
    """
    This class parses GetDP's TimeTable format output files.
    """

    def __init__(self, filePath):
        self.time_values = []
        self.values = []
        # Parse data:
        with open(filePath) as file:
            # If the first line starts with #, we skip it.
            first_line = file.readline()
            if not first_line.startswith("#"):
                number_of_entries = len(
                    re.findall(r"(-?\d+\.?\d*e?[-+]*\d*)", first_line)
                )
                # readline() moves the cursor to the next line, so we need to go back to
                # the beginning of the file.
                file.seek(0)
            else:
                second_line = file.readline()
                number_of_entries = len(
                    re.findall(r"(-?\d+\.?\d*e?[-+]*\d*)", second_line)
                )
                # Seek to the second line
                file.seek(len(first_line) + 1)

            data = file.read()

        entries = re.findall(r"(-?\d+\.?\d*e?[-+]*\d*)", data)
        if number_of_entries == 2:
            # Global scalar value:
            time_index = 0
            value_index = 1
            self.data_type = "scalar"
        elif number_of_entries == 6:
            # Local scalar value probed at a point:
            time_index = 1
            value_index = 5
            self.data_type = "scalar"
        elif number_of_entries == 8:
            # Local vector value probed at a point:
            time_index = 1
            value_index = [5, 6, 7]
            self.data_type = "vector"
        elif number_of_entries == 14:
            # Local tensor value probed at a point:
            time_index = 1
            value_index = [[5, 6, 7], [8, 9, 10], [11, 12, 13]]
            self.data_type = "tensor"
        else:
            raise ValueError(f"{filePath} contains an unexpected type of data.")

        # Pack entries for each line:
        entries = [
            entries[i : i + number_of_entries]
            for i in range(0, len(entries), number_of_entries)
        ]

        for entry in entries:
            if self.data_type == "scalar":
                self.time_values.append(float(entry[time_index]))
                self.values.append(float(entry[value_index]))
            elif self.data_type == "vector":
                self.time_values.append(float(entry[time_index]))
                self.values.append(
                    (
                        float(entry[value_index[0]]),
                        float(entry[value_index[1]]),
                        float(entry[value_index[2]]),
                    )
                )
            elif self.data_type == "tensor":
                self.time_values.append(float(entry[time_index]))
                self.values.append(
                    [
                        [
                            float(entry[value_index[0][0]]),
                            float(entry[value_index[0][1]]),
                            float(entry[value_index[0][2]]),
                        ],
                        [
                            float(entry[value_index[1][0]]),
                            float(entry[value_index[1][1]]),
                            float(entry[value_index[1][2]]),
                        ],
                        [
                            float(entry[value_index[2][0]]),
                            float(entry[value_index[2][1]]),
                            float(entry[value_index[2][2]]),
                        ],
                    ]
                )

    def get_equivalent_scalar_values(self):
        """
        Returns the same scalar if self.data_type is scalar.
        Returns the magnitude of the vectors if self.data_type is vector.
        Returns the von misses equivalents of the tensors if self.data_type is tensor.
        """

        if self.data_type == "scalar":
            return self.values
        elif self.data_type == "vector":
            magnitudes = [
                math.sqrt(v[0] ** 2 + v[1] ** 2 + v[2] ** 2)
                for v in self.values
            ]
            return magnitudes
        elif self.data_type == "tensor":
            von_misses_equivalents = [
                math.sqrt(
                    0.5
                    * (
                        (v[0][0] - v[1][1]) ** 2
                        + (v[1][1] - v[2][2]) ** 2
                        + (v[2][2] - v[0][0]) ** 2
                        + 6
                        * (
                            ((v[0][1] + v[1][0]) / 2) ** 2
                            + ((v[1][2] + v[2][1]) / 2) ** 2
                            + ((v[0][2] + v[2][0]) / 2) ** 2
                        )
                    )
                )
                for v in self.values
            ]
            return von_misses_equivalents
        else:
            raise RuntimeError("Data type not recognized.")



import unittest
import os
import gmsh
import logging
import shutil
from typing import Union, Tuple, Literal
import ruamel.yaml
import numpy as np

from fiqus.data.DataFiQuS import FDM
from fiqus.MainFiQuS import MainFiQuS
from fiqus.utils.Utils import GmshUtils
from fiqus.parsers.ParserMSH import ParserMSH
from fiqus.utils.Utils import FilesAndFolders as FFs

logger = logging.getLogger(__name__)


class BaseClassesForTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Change the working directory to the directory where this file's parent
        # directory is located (tests folder):
        os.chdir(os.path.dirname(os.path.dirname(__file__)))
        cls.inputs_folder = os.path.join(os.getcwd(), "_inputs")
        cls.outputs_folder = os.path.join(os.getcwd(), "_outputs")
        cls.references_folder = os.path.join(os.getcwd(), "_references")

    @classmethod
    def tearDownClass(cls) -> None:
        pass

    def setUp(self) -> None:
        """
        This method is executed before each test in this class.

        It changes the working directory to the directory where this file is located.
        """
        # Change the working directory to the directory where this file's parent
        # directory is located (tests folder):
        os.chdir(os.path.dirname(os.path.dirname(__file__)))

    def tearDown(self) -> None:
        pass

    def get_input_file_path(self, model_name: str) -> Union[str, os.PathLike]:
        """
        This method returns the path to the input file for the given model name.

        :param model_name: name of the model to get the input file for
        :type model_name: str
        :return: path to the input file
        :rtype: Union[str, os.PathLike]
        """
        return os.path.join(self.inputs_folder, f"{model_name}", f"{model_name}.yaml")

    def get_data_model(self, model_name: str) -> FDM:
        """
        This method returns the data model for the given model name by reading the
        input file.

        :param model_name: name of the model to get the data model for
        :type model_name: str
        :return: data model
        :rtype: FDM
        """
        input_file_path = self.get_input_file_path(model_name)

        # Read and prepare the input file:
        fdm = FFs.read_data_from_yaml(input_file_path, FDM)

        if fdm.run.geometry is None:
            fdm.run.geometry = 1
        if fdm.run.mesh is None:
            fdm.run.mesh = 1
        if fdm.run.solution is None:
            fdm.run.solution = 1

        return fdm

    def run_fiqus(
        self,
        data_model: FDM,
        model_name: str,
        run_type: Literal[
            "start_from_yaml",
            "mesh_only",
            "geometry_only",
            "geometry_and_mesh",
            "pre_process_only",
            "mesh_and_solve_with_post_process_python",
            "solve_with_post_process_python",
            "solve_only",
            "post_process_getdp_only",
            "post_process_python_only",
        ],
    ) -> None:
        """
        This method runs FiQuS with the given model name and run type.

        :param model_name: name of the model to run FiQuS for
        :type model_name: str
        :param run_type: run type to run FiQuS with
        :type run_type: Literal[
                "start_from_yaml",
                "mesh_only",
                "geometry_only",
                "geometry_and_mesh",
                "pre_process_only",
                "mesh_and_solve_with_post_process_python",
                "solve_with_post_process_python",
                "solve_only",
                "post_process_getdp_only",
                "post_process_python_only",
            ]
        """
        # Make sure the run type is correct:
        data_model.run.type = run_type

        # Make sure the output files are overwritten:
        data_model.run.overwrite = True

        # Prepare the output folder:
        model_folder = os.path.join(self.outputs_folder, f"{model_name}_{run_type}")
        FFs.prep_folder(model_folder)

        # Depending on the run_type, copy the reference files to the model folder:
        if run_type in ["geometry_only", "geometry_and_mesh", "start_from_yaml"]:
            pass
        else:
            # Copy the reference geometry and mesh folders to the model folder:
            reference_geometry_folder = self.get_path_to_specific_reference_folder(
                model_name, folder="Geometry"
            )
            reference_mesh_folder = self.get_path_to_specific_reference_folder(
                model_name, folder="Mesh"
            )

            output_geometry_folder = self.get_path_to_specific_output_folder(
                model_name, run_type=run_type, folder="Geometry"
            )
            output_mesh_folder = self.get_path_to_specific_output_folder(
                model_name, run_type=run_type, folder="Mesh"
            )

            shutil.copytree(
                reference_geometry_folder, output_geometry_folder, dirs_exist_ok=True
            )
            shutil.copytree(
                reference_mesh_folder, output_mesh_folder, dirs_exist_ok=True
            )

        # Run FiQuS:
        MainFiQuS(
            model_folder=model_folder,
            input_file_path=self.get_input_file_path(model_name),
            fdm=data_model,
            verbose=False,
        )

    def get_path_to_generated_file(
        self,
        model_name: str,
        file_name: str,
        file_extension: str,
        run_type: Literal[
            "start_from_yaml",
            "mesh_only",
            "geometry_only",
            "geometry_and_mesh",
            "pre_process_only",
            "mesh_and_solve_with_post_process_python",
            "solve_with_post_process_python",
            "solve_only",
            "post_process_getdp_only",
            "post_process_python_only",
        ],
    ) -> Union[str, os.PathLike]:
        """
        This method returns the path to the generated file with the given extension
        for the given model name depending on the run type.
        
        :param model_name: name of the model to get the generated file for
        :type model_name: str
        :param file_name: name of the file
        :type file_name: str
        :param file_extension: file extension of the generated file
        :type file_extension: str
        :param run_type: run type to run FiQuS with
        :type run_type: Literal[
                "start_from_yaml",
                "mesh_only",
                "geometry_only",
                "geometry_and_mesh",
                "pre_process_only",
                "mesh_and_solve_with_post_process_python",
                "solve_with_post_process_python",
                "solve_only",
                "post_process_getdp_only",
                "post_process_python_only",
            ]
        :return: path to the generated file
        :rtype: Union[str, os.PathLike]
        """
        if run_type == "geometry_only":
            folder = "Geometry"
        elif run_type == "mesh_only":
            folder = "Mesh"
        elif run_type == "solve_only":
            folder = "Solution"
        else:
            raise ValueError(
                "The run type must be geometry_only, mesh_only, or solve_only!"
            )

        sections_folder = self.get_path_to_specific_output_folder(
            model_name, run_type, folder
        )

        generated_file = os.path.join(
            sections_folder,
            f"{file_name}.{file_extension}",
        )

        # Check if the file exists:
        if not os.path.isfile(generated_file):
            raise FileNotFoundError(
                f"Could not find the generated file: {generated_file}!"
            )

        return generated_file

    def get_path_to_specific_reference_folder(
        self,
        model_name: str,
        folder: Literal[
            "Geometry",
            "Mesh",
            "Solution",
        ],
    ):
        """
        This method returns the path to a specific reference folder (Geometry, Mesh, or
        Solution) for the given model name.

        :param model_name: name of the model to get the reference folder for
        :type model_name: str
        :param folder: folder to get the reference folder for
        :type folder: Literal[
                "Geometry",
                "Mesh",
                "Solution",
            ]
        """
        fdm = self.get_data_model(model_name)
        geometry_folder_name = f"Geometry_{fdm.run.geometry}"
        mesh_folder_name = f"Mesh_{fdm.run.mesh}"
        solve_folder_name = f"Solution_{fdm.run.solution}"

        if folder == "Geometry":
            reference_folder = os.path.join(
                self.references_folder, model_name, geometry_folder_name
            )
        elif folder == "Mesh":
            reference_folder = os.path.join(
                self.references_folder,
                model_name,
                geometry_folder_name,
                mesh_folder_name,
            )
        elif folder == "Solution":
            reference_folder = os.path.join(
                self.references_folder,
                model_name,
                geometry_folder_name,
                mesh_folder_name,
                solve_folder_name,
            )

        return reference_folder

    def get_path_to_specific_output_folder(
        self,
        model_name: str,
        run_type: Literal[
            "start_from_yaml",
            "mesh_only",
            "geometry_only",
            "geometry_and_mesh",
            "pre_process_only",
            "mesh_and_solve_with_post_process_python",
            "solve_with_post_process_python",
            "solve_only",
            "post_process_getdp_only",
            "post_process_python_only",
        ],
        folder: Literal[
            "Geometry",
            "Mesh",
            "Solution",
        ],
    ):
        """
        This method returns a specific path (Geometry, Mesh, or Solution) to the output
        folder for the given model name depending on the run type.

        :param model_name: name of the model to get the output folder for
        :type model_name: str
        :param run_type: run type to run FiQuS with
        :type run_type: Literal[
                "start_from_yaml",
                "mesh_only",
                "geometry_only",
                "geometry_and_mesh",
                "pre_process_only",
                "mesh_and_solve_with_post_process_python",
                "solve_with_post_process_python",
                "solve_only",
                "post_process_getdp_only",
                "post_process_python_only",
            ]
        :param folder: folder to get the output folder for
        :type folder: Literal[
                "Geometry",
                "Mesh",
                "Solution",
            ]
        """
        fdm = self.get_data_model(model_name)
        geometry_folder_name = f"Geometry_{fdm.run.geometry}"
        mesh_folder_name = f"Mesh_{fdm.run.mesh}"
        solve_folder_name = f"Solution_{fdm.run.solution}"

        if run_type is None:
            model_folder_name = model_name
        else:
            model_folder_name = f"{model_name}_{run_type}"

        if folder == "Geometry":
            output_folder = os.path.join(
                self.outputs_folder, model_folder_name, geometry_folder_name
            )
        elif folder == "Mesh":
            output_folder = os.path.join(
                self.outputs_folder,
                model_folder_name,
                geometry_folder_name,
                mesh_folder_name,
            )
        elif folder == "Solution":
            output_folder = os.path.join(
                self.outputs_folder,
                model_folder_name,
                geometry_folder_name,
                mesh_folder_name,
                solve_folder_name,
            )

        FFs.prep_folder(output_folder)

        return output_folder

    def get_path_to_reference_file(
        self,
        model_name: str,
        file_name: str,
        file_extension: str,
        folder: Literal[
            "Geometry",
            "Mesh",
            "Solution",
        ],
    ) -> Union[str, os.PathLike]:
        """
        This method returns the path to the reference file with the given extension
        for the given model name from the given folder.
        
        :param model_name: name of the model to get the reference file for
        :type model_name: str
        :param file_name: name of the model to get the reference file for
        :type file_name: str
        :param file_extension: file extension of the reference file
        :type file_extension: str
        :param folder: folder to get the reference file from
        :type folder: Literal[
                "Geometry",
                "Mesh",
                "Solution",
            ]
        :return: path to the reference file
        :rtype: Union[str, os.PathLike]
        """
        reference_folder = self.get_path_to_specific_reference_folder(
            model_name, folder=folder
        )
        reference_file = os.path.join(
            reference_folder,
            f"{file_name}.{file_extension}",
        )

        # Check if the file exists:
        if not os.path.isfile(reference_file):
            raise FileNotFoundError(
                f"Could not find the reference file: {file_name}.{file_extension}!"
            )

        return reference_file

    def compare_json_or_yaml_files(self, file_1, file_2):
        """
        This method compares the contents of two JSON or YAML files. It is used to
        check that the generated files are the same as the reference.

        :param file_1: path to the first file
        :type file_1: Union[str, os.PathLike]
        :param file_2: path to the second file
        :type file_2: Union[str, os.PathLike]
        """
        try:
            # YAML is a superset of JSON, so we can use the same parser for both:
            yaml = ruamel.yaml.YAML(typ="safe", pure=True)
            with open(file_1, "r") as file:
                file_1_dictionary = yaml.load(file)

            with open(file_2, "r") as file:
                file_2_dictionary = yaml.load(file)
        except:
            raise ValueError("The files must be JSON or YAML files!")

        # Compare the dictionaries:
        self.assertDictEqual(
            file_1_dictionary,
            file_2_dictionary,
            msg=f"{os.path.basename(file_1)} did not match {os.path.basename(file_2)}!",
        )

    def compare_text_files(self, file_1, file_2):
        """
        This method compares the contents of two text files. It is used to check that
        the generated files are the same as the reference.

        :param file_1: path to the first file
        :type file_1: Union[str, os.PathLike]
        :param file_2: path to the second file
        :type file_2: Union[str, os.PathLike]
        """
        # Open the files and get the contents:
        with open(file_1, "r") as file:
            file_1_contents = file.read()

        with open(file_2, "r") as file:
            file_2_contents = file.read()

        # Compare the contents:
        self.assertEqual(
            file_1_contents,
            file_2_contents,
            msg=f"{os.path.basename(file_1)} did not match {os.path.basename(file_2)}!",
        )


class FiQuSGeometryTests(BaseClassesForTests):
    def generate_geometry(
        self,
        data_model: FDM,
        model_name: str,
    ) -> Tuple[Union[str, os.PathLike], Union[str, os.PathLike]]:
        """
        This method generates the geometry for the given model name.

        :param data_model: data model to run FiQuS with
        :type data_model: FDM
        :param model_name: name of the model to generate the geometry for
        :type model_name: str
        :return: path to the generated geometry and volume information files
        :rtype: Tuple[Union[str, os.PathLike], Union[str, os.PathLike]]
        """
        self.run_fiqus(data_model, model_name, "geometry_only")
        self.model_name = model_name

    def compare_number_of_entities(
        self,
        geometry_file_1: Union[str, os.PathLike],
        geometry_file_2: Union[str, os.PathLike],
    ):
        """
        This method compares the number of entities for each dimension in two geometry
        files. It is used to check that the generated geometry file is the same as the
        reference.

        :param geometry_file_1: path to the first geometry file
        :type geometry_file_1: Union[str, os.PathLike]
        :param geometry_file_2: path to the second geometry file
        :type geometry_file_2: Union[str, os.PathLike]
        """
        # Initialize gmsh:
        gmsh_utils = GmshUtils(verbose=False)
        gmsh_utils.initialize()

        # Open the geometry files and get the entities:
        model_entities = {}
        for geometry_file in [geometry_file_1, geometry_file_2]:
            model_entities[geometry_file] = []

            gmsh.open(geometry_file)
            for dim in range(4):
                model_entities[geometry_file].append(
                    sorted(gmsh.model.getEntities(dim=dim))
                )
            gmsh.clear()

        # Compare the number of entities for each dimension:
        for dim in [0, 1, 2, 3]:
            self.assertEqual(
                len(model_entities[geometry_file_1][dim]),
                len(model_entities[geometry_file_2][dim]),
                msg=f"{geometry_file_1} and {geometry_file_2} did not match!",
            )

    def get_path_to_generated_file(
        self, file_name: str, file_extension: str
    ) -> Union[str, os.PathLike]:
        return super().get_path_to_generated_file(
            self.model_name, file_name, file_extension, run_type="geometry_only"
        )

    def get_path_to_reference_file(
        self, file_name: str, file_extension: str
    ) -> Union[str, os.PathLike]:
        return super().get_path_to_reference_file(
            self.model_name, file_name, file_extension, folder="Geometry"
        )


class FiQuSMeshTests(BaseClassesForTests):
    def generate_mesh(
        self,
        data_model: FDM,
        model_name: str,
    ) -> Tuple[Union[str, os.PathLike], Union[str, os.PathLike]]:
        """
        This method generates the mesh for the given model name.

        :param data_model: data model to run FiQuS with
        :type data_model: FDM
        :param model_name: name of the model to generate the mesh for
        :type model_name: str
        """
        self.run_fiqus(data_model, model_name, "mesh_only")
        self.model_name = model_name

    def compare_mesh_qualities(
        self, mesh_file_1: Union[str, os.PathLike], mesh_file_2: Union[str, os.PathLike]
    ):
        """
        This method compares the mesh qualities of two mesh files. It is used to check
        that the generated mesh file is the same as the reference.

        :param mesh_file_1: path to the first mesh file
        :type mesh_file_1: Union[str, os.PathLike]
        :param mesh_file_2: path to the second mesh file
        :type mesh_file_2: Union[str, os.PathLike]
        """
        # Initialize gmsh:
        gmsh_utils = GmshUtils(verbose=False)
        gmsh_utils.initialize()

        # Open the mesh files and get the average mesh quality:
        average_mesh_qualities = []
        for mesh_file in [mesh_file_1, mesh_file_2]:
            # Get physical group names:
            mesh = ParserMSH(mesh_file)
            average_mesh_qualities.append(mesh.get_average_mesh_quality())

        # Make sure the average mesh qualities are close enough:
        ratio = min(average_mesh_qualities) / max(average_mesh_qualities)
        self.assertGreater(
            ratio,
            0.9,
            msg=f"{mesh_file_1} and {mesh_file_2} did not match!",
        )

    def get_path_to_generated_file(
        self, file_name: str, file_extension: str
    ) -> Union[str, os.PathLike]:
        return super().get_path_to_generated_file(
            self.model_name, file_name, file_extension, run_type="mesh_only"
        )

    def get_path_to_reference_file(
        self, file_name: str, file_extension: str
    ) -> Union[str, os.PathLike]:
        return super().get_path_to_reference_file(
            self.model_name, file_name, file_extension, folder="Mesh"
        )


class FiQuSSolverTests(BaseClassesForTests):
    def solve(
        self,
        data_model: FDM,
        model_name: str,
    ) -> Tuple[Union[str, os.PathLike], Union[str, os.PathLike]]:
        """
        This method solves the given model name.

        :param data_model: data model to run FiQuS with
        :type data_model: FDM
        :param model_name: name of the model to generate the mesh for
        :type model_name: str
        """
        self.run_fiqus(data_model, model_name, "solve_only")
        self.model_name = model_name

    def compare_pos_files(
        self, pos_file_1: Union[str, os.PathLike], pos_file_2: Union[str, os.PathLike]
    ):
        """
        This method compares the contents of two pos files. It is used to check that
        the generated pos file is the same as the reference.

        :param pos_file_1: path to the first pos file
        :type pos_file_1: Union[str, os.PathLike]
        :param pos_file_2: path to the second pos file
        :type pos_file_2: Union[str, os.PathLike]
        """
        # Initialize gmsh:
        gmsh_utils = GmshUtils(verbose=False)
        gmsh_utils.initialize()

        # Open the pos files and get the model data:
        model_datas = []
        for pos_file in [pos_file_1, pos_file_2]:
            # Open the pos file:
            gmsh.open(pos_file)
            for time_step in range(1, 100):
                data_all_steps = []
                # Save all available time steps up to 100:
                try:
                    (
                        data_type,
                        tags,
                        data,
                        time,
                        numComponents,
                    ) = gmsh.view.getHomogeneousModelData(tag=0, step=time_step)
                    data_all_steps.extend(data)
                except:
                    break

            model_datas.append(data_all_steps)

        # Make sure the pos files are close enough:
        with self.subTest(msg=f"{pos_file_1} and {pos_file_2} did not match!"):
            self.assertEqual(
                len(model_datas[0]),
                len(model_datas[1]),
            )

            # Convert to numppy array:
            model_data1 = np.array(model_datas[0])
            model_data2 = np.array(model_datas[1])

            # Compare the data:
            np.testing.assert_allclose(model_data1, model_data2, rtol=1e-05, atol=1e-08)

    def get_path_to_generated_file(
        self, file_name: str, file_extension: str
    ) -> Union[str, os.PathLike]:
        return super().get_path_to_generated_file(
            self.model_name, file_name, file_extension, run_type="solve_only"
        )

    def get_path_to_reference_file(
        self, file_name, file_extension
    ) -> Union[str, os.PathLike]:
        return super().get_path_to_reference_file(
            self.model_name, file_name, file_extension, folder="Solution"
        )



import unittest
import os
from pathlib import Path
import gmsh
import numpy as np
import pandas as pd
import getpass

from fiqus.data.DataFiQuS import FDM
from fiqus.MainFiQuS import MainFiQuS
from fiqus.parsers.ParserMSH import ParserMSH
from fiqus.parsers.ParserPOS import ParserPOS
import tests.utils.helpers as tuh
from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import FilesAndFolders as FFs
from fiqus.utils.Utils import RoxieParsers as Pars


class TestFiQuS(unittest.TestCase):

    def setUp(self) -> None:
        """
            This function is executed before each test in this class
        """
        self.current_path = os.getcwd()
        os.chdir(os.path.dirname(__file__))  # move to the directory where this file is located
        print('Test is run from folder: {}'.format(os.getcwd()))

        self.model_names = ['MQXA', 'MCBRD_2d2a_2n2a_0i']

    def tearDown(self) -> None:
        """
            This function is executed after each test in this class
        """
        os.chdir(self.current_path)  # go back to initial folder

    def test_01_GenerateAllGeometryOnly(self):
        """
            Check that geometry file (.brep) can be generated and is the same as the reference.
        """
        # arrange
        for model_name in self.model_names:
            p = tuh.Paths(model_name, 'brep')
            FFs.prep_folder(p.test_outputs_folder)

            fdm = FFs.read_data_from_yaml(p.input_file, FDM)
            fdm.run.type = 'geometry_only'
            fdm.run.geometry = 1
            fdm.run.overwrite = True

            # act
            main_fiqus = MainFiQuS(input_file_path=p.input_file, model_folder=p.model_folder, fdm=fdm)

            # assert
            gu = GmshUtils(model_name, True)
            gu.initialize()

            print('Comparing model and reference .brep files')
            print(f'Model file: {main_fiqus.main_magnet.model_file}')
            gmsh.open(main_fiqus.main_magnet.model_file)
            model_entities = []
            for dim in range(4):
                model_entities.append(sorted(gmsh.model.getEntities(dim=dim)))
            gmsh.clear()

            print(f'Reference file: {p.reference_file}')
            gmsh.open(p.reference_file)
            reference_entities = []
            for dim in range(4):
                reference_entities.append(sorted(gmsh.model.getEntities(dim=dim)))
            gmsh.clear()

            self.assertEqual(len(model_entities), len(reference_entities))
            for model_ents, reference_ents in zip(model_entities, reference_entities):
                self.assertEqual(len(model_ents), len(reference_ents))
                for model_ent, reference_ent in zip(model_ents, reference_ents):
                    self.assertEqual(model_ent[1], reference_ent[1], msg=f'Models elements are not equal for dimension: {reference_ent[0]}!')
            print('The number of points, lines, surfaces and volumes in the .brep files match!')

    def test_02_GenerateAllMeshOnly(self):
        """
            Check that mesh file with correct physical names gets generated
        """
        # arrange
        for model_name in self.model_names:
            p = tuh.Paths(model_name, 'msh')
            FFs.prep_folder(p.test_outputs_folder)

            fdm = FFs.read_data_from_yaml(p.input_file, FDM)
            fdm.run.type = 'mesh_only'
            fdm.run.geometry = 1
            fdm.run.mesh = 1
            fdm.run.overwrite = True

            # act
            main_fiqus = MainFiQuS(input_file_path=p.input_file, model_folder=p.model_folder, fdm=fdm)

            # assert
            # this only compares if physical group names in the mesh files are equal. i.e. are all regions created, including cuts in a correct way.
            print('Comparing physical group names in the output and reference .msh files')
            print(f'Model file: {main_fiqus.main_magnet.model_file}')
            print(f'Reference file: {p.reference_file}')
            output_physical_names = ParserMSH(main_fiqus.main_magnet.model_file).physical_names
            reference_physical_names = ParserMSH(p.reference_file).physical_names
            print(f'Output physical group names:'
                  f'{output_physical_names}')
            print(f'Reference physical group names:'
                  f'{reference_physical_names}')
            self.assertEqual(output_physical_names, reference_physical_names)
            print('The mesh files physical group names match!')

    def test_03_AllSolveOnly(self):
        """
            Check that Solution pos file is identical to the reference one
        """
        # arrange
        for model_name in self.model_names:
            p = tuh.Paths(model_name, '')
            FFs.prep_folder(p.test_outputs_folder)

            fdm = FFs.read_data_from_yaml(p.input_file, FDM)
            fdm.run.type = 'solve_only'
            fdm.run.geometry = 1
            fdm.run.mesh = 1
            fdm.run.solution = 1
            fdm.run.overwrite = True
            # fdm.run.launch_gui = True

            # act
            main_fiqus = MainFiQuS(input_file_path=p.input_file, model_folder=p.model_folder, fdm=fdm)

            # assert
            columns = ['x', 'y', 'z', 'NaN1', 'NaN2', 'Bx', 'By', 'Bz', 'NaN3', 'NaN4']
            delimiter = ' '
            print('Comparing magnetic field csv files')
            model_file = main_fiqus.main_magnet.model_file
            print(f'Output file: {model_file}')
            model_df = pd.read_csv(model_file, delimiter=delimiter, header=None, engine='python', names=columns)
            reference_file = os.path.join(p.references_folder, 'Center_line.csv')
            print(f'Reference file: {reference_file}')
            reference_df = pd.read_csv(reference_file, delimiter=delimiter, header=None, engine='python', names=columns)

            coord_atol = 1e-9
            field_atol = 1e-3
            comparison_atol = [coord_atol, coord_atol, coord_atol, None, None, field_atol, field_atol, field_atol, None, None]
            for column, atol in zip(columns, comparison_atol):
                if atol:
                    print(f'Comparing {column} values with absolute tolerance of {atol}')
                    comparison_result = np.isclose(np.mean(model_df[column]), np.mean(reference_df[column]), rtol=0, atol=atol, equal_nan=True).all()
                    self.assertTrue(comparison_result)
                    print(f'Data in {column} are within tolerance')
            print('The magnetic field values in csv files satisfy test criteria')

    def test_04_PostProcessCCTGetDPOnly(self):
        """
            Check that post processing step defined in the .pro file can be run in GetDP. This is helper tests for debugging postprocessing step of GetDP
        """
        # arrange
        for model_name in self.model_names:
            p = tuh.Paths(model_name, '')
            FFs.prep_folder(p.test_outputs_folder)

            fdm = FFs.read_data_from_yaml(p.input_file, FDM)
            if fdm.magnet.type == 'CCT_straight':
                fdm.run.type = 'post_process_getdp_only'
                fdm.run.geometry = 1
                fdm.run.mesh = 1
                fdm.run.solution = 1

                # act
                main_fiqus = MainFiQuS(input_file_path=p.input_file, model_folder=p.model_folder, fdm=fdm)

                # assert
                model_df = pd.read_csv(os.path.join(main_fiqus.main_magnet.model_folder, 'Magnetic_energy.dat'))
                print(f"Magnetic_energy: {model_df}")
                model_df = pd.read_csv(os.path.join(main_fiqus.main_magnet.model_folder, 'Inductance.dat'))
                print(f"Inductance: {model_df}")

    def test_05_PostProcessMultipoleGetDPOnly(self):
        pass

    def test_06_PostProcessCCTPythonOnly(self):
        """
            Check that post processing of getdp ouput .pos file with magnetic field by python into csv file for LEDET is identical to the reference one
        """
        # arrange
        for model_name in self.model_names:
            p = tuh.Paths(model_name, '')
            FFs.prep_folder(p.test_outputs_folder)

            fdm = FFs.read_data_from_yaml(p.input_file, FDM)
            if fdm.magnet.type == 'CCT_straight':
                fdm.run.type = 'post_process_python_only'
                fdm.run.geometry = 1
                fdm.run.mesh = 1
                fdm.run.solution = 1

                # act
                main_fiqus = MainFiQuS(input_file_path=p.input_file, model_folder=p.model_folder, fdm=fdm)

                # assert
                print('Comparing magnetic field csv files')
                model_file = main_fiqus.main_magnet.model_file
                print(f'Model file: {model_file}')
                model_df = pd.read_csv(model_file)
                reference_file = os.path.join(p.references_folder, 'field_map_3D.csv')
                print(f'Reference file: {reference_file}')
                reference_df = pd.read_csv(reference_file)

                for column in reference_df:
                    if column in ['Bx [T]', 'By [T]', 'Bz [T]', 'Bl [T]', 'Bh [T]', 'Bw [T]']:
                        comp_abs_tol = 5e-3  # this is comparing averages, not very robust and needs improvement
                        print(f'Comparing {column} average absolute tolerance of {comp_abs_tol}')
                        comparison_result = np.isclose(model_df[column].mean(), reference_df[column].mean(), rtol=0, atol=comp_abs_tol, equal_nan=True)
                    else:
                        comp_abs_tol = 1e-9
                        print(f'Comparing {column} values with absolute tolerance of {comp_abs_tol}')
                        comparison_result = np.isclose(model_df[column], reference_df[column], rtol=0, atol=comp_abs_tol, equal_nan=True).all()
                    self.assertTrue(comparison_result)
                    print(f'Data in {column} are within tolerance')
                print('The magnetic field values in csv files satisfy test criteria')

                # assert        # this is done to improve coverage by testing ParserPOS. Test only if ParserPOS does not produce error.
                file_name = 'b_Omega_p.pos'
                model_file = os.path.join(main_fiqus.main_magnet.model_folder, file_name)
                print(f'Parsing file: {model_file}')
                ParserPOS(model_file)
                print(f'Parsing of {model_file} with ParserPOS was successful.')

    def test_07_PostProcessMultipolePythonOnly(self):
        """
            Check that post processing csv/map2d file for LEDET is identical to the reference one
        """
        # arrange
        for model_name in self.model_names:
            p = tuh.Paths(model_name, '')
            FFs.prep_folder(p.test_outputs_folder)

            fdm = FFs.read_data_from_yaml(p.input_file, FDM)
            if fdm.magnet.type == 'multipole':
                fdm.run.type = 'post_process_python_only'
                fdm.run.geometry = 1
                fdm.run.mesh = 1
                fdm.run.solution = 1

                # act
                main_fiqus = MainFiQuS(input_file_path=p.input_file, model_folder=p.model_folder, fdm=fdm)

                # assert
                p.reference_file = Path(f"{p.reference_file}map2d")
                print('Comparing map2d files')
                reference_map2d = Pars.parseMap2d(map2dFile=p.reference_file)
                output_map2d = Pars.parseMap2d(map2dFile=main_fiqus.main_magnet.model_file)
                comp_abs_tol = 5e-2
                for i, row in enumerate(reference_map2d):
                    print(f'Comparing values of row {i} with absolute tolerance of {comp_abs_tol}')
                    for j, entry in enumerate(row):
                        comparison_result = np.isclose(entry, output_map2d[i][j], rtol=0, atol=comp_abs_tol, equal_nan=True)
                        self.assertTrue(comparison_result)
                print('The magnetic field values in map2d files satisfy test criteria')

    def test_08_PostProcessMultipolePythonOnlyCompRoxie(self):
        """
            Check that post processing csv/map2d file for LEDET is identical to the reference one from ROXIE
        """
        # arrange
        for model_name in self.model_names:
            p = tuh.Paths(model_name, 'map2d')
            FFs.prep_folder(p.test_outputs_folder)

            fdm = FFs.read_data_from_yaml(p.input_file, FDM)

            if fdm.magnet.type == 'multipole':
                fdm.run.type = 'post_process_python_only'
                fdm.run.geometry = 1
                fdm.run.mesh = 1
                fdm.run.solution = 1
                fdm.magnet.postproc.compare_to_ROXIE = os.path.join(os.path.dirname(p.input_file), os.path.basename(p.model_file))

                # act
                main_fiqus = MainFiQuS(input_file_path=p.input_file, model_folder=p.model_folder, fdm=fdm)

                # assert
                print('Comparing map2d files')
                reference_map2d = Pars.parseMap2d(map2dFile=Path(os.path.join(p.references_folder, model_name + '_ROXIE.map2d')))
                output_map2d = Pars.parseMap2d(map2dFile=main_fiqus.main_magnet.model_file)
                comp_abs_tol = 5e-2
                for i, row in enumerate(reference_map2d):
                    print(f'Comparing values of row {i} with absolute tolerance of {comp_abs_tol}')
                    for j, entry in enumerate(row):
                        comparison_result = np.isclose(entry, output_map2d[i][j], rtol=0, atol=comp_abs_tol, equal_nan=True)
                        self.assertTrue(comparison_result)
                print('The magnetic field values in map2d files satisfy test criteria')

    def test_09_PostProcessMultipolePythonOnlyPlots(self):
        """
            Check that plots are produced correctly
        """
        # arrange
        for model_name in self.model_names:
            p = tuh.Paths(model_name, 'map2d')
            FFs.prep_folder(p.test_outputs_folder)

            fdm = FFs.read_data_from_yaml(p.input_file, FDM)

            if fdm.magnet.type == 'multipole':
                fdm.run.type = 'post_process_python_only'
                fdm.run.geometry = 1
                fdm.run.mesh = 1
                fdm.run.solution = 1
                fdm.magnet.postproc.compare_to_ROXIE = os.path.join(os.path.dirname(p.input_file), os.path.basename(p.model_file))
                fdm.magnet.postproc.plot_all = 'test'

                # act and assert
                MainFiQuS(input_file_path=p.input_file, model_folder=p.model_folder, fdm=fdm)
                print('The plotting functions satisfy test criteria')

    def test_10_TestGUILaunch(self):
        """
            Check if interactive gui can be launched
        """
        print('Initializing gmsh and launching graphical user interface. It will be open for 0.5 s')
        gu = GmshUtils('test', True)
        gu.initialize()
        user_name = getpass.getuser()
        if user_name != 'root':
            gu.launch_interactive_GUI(close_after=0.5)
        print('Gui was launched and closed successfully')








import sys
import os
import shutil
import numpy as np
from pathlib import Path
from time import sleep
import multiprocessing
import ruamel.yaml
import warnings
import gmsh
import logging
import re

logger = logging.getLogger(__name__)

class LoggingFormatter(logging.Formatter):
    """
    Logging formatter class
    """
    grey = "\x1b[38;20m" # debug level
    white = "\x1b[37;20m" # info level
    yellow = "\x1b[33;20m" # warning level
    red = "\x1b[31;20m" # error level
    bold_red = "\x1b[31;1m" # critical level

    reset = "\x1b[0m"
    format = '%(asctime)s | %(levelname)s | %(message)s'

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: white + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class FilesAndFolders:
    @staticmethod
    def read_data_from_yaml(full_file_path, data_class):
        with open(full_file_path, 'r') as stream:
            yaml = ruamel.yaml.YAML(typ='safe', pure=True)
            yaml_str = yaml.load(stream)
        return data_class(**yaml_str)

    @staticmethod
    def write_data_to_yaml(full_file_path, dict_of_data_class):
        def my_represent_none(self, data):
            return self.represent_scalar('tag:yaml.org,2002:null', 'null')

        yaml = ruamel.yaml.YAML()
        yaml.default_flow_style = False
        yaml.emitter.alt_null = 'Null'
        yaml.representer.add_representer(type(None), my_represent_none)
        with open(full_file_path, 'w') as yaml_file:
            yaml.dump(dict_of_data_class, yaml_file)

    @staticmethod
    def prep_folder(folder_full_path, clear: bool = False):
        if clear:
            if os.path.exists(folder_full_path):
                shutil.rmtree(folder_full_path)  # delete directory
        if not os.path.exists(folder_full_path):
            os.makedirs(folder_full_path)  # make new directory

    @staticmethod
    def get_folder_path(folder_type, folder, ref_nr, overwrite, required_folder):
        if required_folder and not (ref_nr and overwrite):
            last_nr = 0
            for study in [x.parts[-1] for x in Path(folder).iterdir() if x.is_dir()]:
                last_nr = max(int(study[study.find('_') + 1:]), last_nr)
            if overwrite and required_folder and last_nr > 0:
                run_nr = str(last_nr)
            else:
                run_nr = str(last_nr + 1)
        else:
            run_nr = str(ref_nr)

        folder_path = os.path.join(folder, folder_type + '_' + run_nr)
        FilesAndFolders.prep_folder(folder_path, overwrite and required_folder)
        return folder_path

    @staticmethod
    def get_folder_path(folder_type, folder, folder_key, overwrite, required_folder):
        if required_folder and not (folder_key and overwrite):
            all_dirs = [x.parts[-1] for x in Path(folder).iterdir() if x.is_dir()]
            all_relevant_dirs = [x for x in all_dirs if x.startswith(f"{folder_type}_{folder_key}")]
            if f"{folder_type}_{folder_key}" in all_relevant_dirs:
                new_folder_key = f"{folder_key}_{len(all_relevant_dirs) + 1}"
                folder_key = new_folder_key

        folder_path = os.path.join(folder, folder_type + '_' + str(folder_key))
        # Disable the line below to avoid deleating volder
        FilesAndFolders.prep_folder(folder_path, overwrite and required_folder)
        return folder_path

    @staticmethod
    def compute_folder_key(folder_type, folder, overwrite):
        # Find all the directories in the folder
        all_dirs = [x.parts[-1] for x in Path(folder).iterdir() if x.is_dir()]

        # Find all the directiories that start with the folder_type (e.g. geometry, mesh, solution)
        # Then combine them into a single string with a custom seperator (se@p)
        # Seperators are used to guarantee the directories can be split later
        all_relevant_dirs = " se@p ".join([x for x in all_dirs if x.startswith(f"{folder_type}_")])
        all_relevant_dirs = f"{all_relevant_dirs} se@p "

        # Find all the integer keys in the relevant directories
        integers_in_relevant_dirs = re.findall(rf'{folder_type}_(\d+) se@p ', all_relevant_dirs)

        if integers_in_relevant_dirs is None:
            # If there are no integers in the relevant directories, set the key to 1
            folder_key = 1
        else:
            # Make a list of integers out of the integers in the relevant directories
            integers_in_relevant_dirs = [int(x) for x in integers_in_relevant_dirs]

            # Sort the integers in the relevant directories
            integers_in_relevant_dirs.sort()

            if overwrite:
                # If overwrite is true, set the key to the largest integer in the
                # so that the folder with the largest integer key is overwritten
                if len(integers_in_relevant_dirs) == 0:
                    folder_key = 1
                else:
                    folder_key = max(integers_in_relevant_dirs)
            else:
                # If overwrite is false, then find the smallest integer key that is not
                # in the list of integers in the relevant directories
                folder_key = 1
                for i in integers_in_relevant_dirs:
                    if folder_key < i:
                        break
                    folder_key += 1
        
        return folder_key

    @staticmethod
    def print_welcome_graphics():
        print(r"  _____ _  ___        ____ ")
        print(r"|  ___(_)/ _ \ _   _/ ___| ")
        print(r"| |_  | | | | | | | \___ \ ")
        print(r"|  _| | | |_| | |_| |___) |")
        print(r"|_|   |_|\__\_\\__,_|____/ ")
        print("")


class CheckForExceptions:
    @staticmethod
    def check_inputs(run):  # RunFiQuS()
        if run.type == 'start_from_yaml':
            if run.geometry and not run.overwrite:
                warnings.warn("Warning: Geometry folder is needed only if it has to be overwritten. Ignoring it...")
            if run.solution or run.mesh:
                warnings.warn("Warning: Mesh and Solution folders are not needed. Ignoring them...")
        elif run.type == 'geometry_only':
            if run.solution or run.mesh:
                warnings.warn("Warning: Mesh and Solution folders are not needed. Ignoring them...")
        elif run.type == 'mesh_and_solve_with_post_process':
            if not run.geometry:
                raise Exception('Full path to Geometry not provided. '
                                'Insert options -> reference_files -> geometry.')
            if run.mesh and not run.overwrite:
                warnings.warn("Warning: Mesh folder is needed only if it has to be overwritten. Ignoring it...")
            if run.solution:
                warnings.warn("Warning: Solution folder is not needed. Ignoring it...")
        elif run.type == 'mesh_only':
            if not run.geometry:
                raise Exception('Full path to Mesh not provided. '
                                'Insert options -> reference_files -> geometry.')
            if run.solution:
                warnings.warn("Warning: Solution folder is not needed. Ignoring it...")
        elif run.type == 'solve_with_post_process':
            if not run.mesh or not run.geometry:
                raise Exception('Full path to Mesh not provided. '
                                'Insert options -> reference_files -> geometry and mesh.')
            if run.solution and not run.overwrite:
                warnings.warn("Warning: Solution folder is needed only if it has to be overwritten. Ignoring it...")
        elif run.type == 'solve_only':
            if not run.mesh or not run.geometry:
                raise Exception('Full path to Mesh not provided. '
                                'Insert options -> reference_files -> geometry and mesh.')
            if run.solution and not run.overwrite:
                warnings.warn("Warning: Solution folder is needed only if it has to be overwritten. Ignoring it...")
        elif run.type == 'post_process_only':
            if not run.mesh or not run.geometry or not run.solution:
                raise Exception('Full path to Solution not provided. '
                                'Insert options -> reference_files -> geometry, mesh, and solution.')

    @staticmethod
    def check_overwrite_conditions(folder_type, folder, folder_key):
        if folder_key:
            if not os.path.exists(os.path.join(folder, folder_type + '_' + str(folder_key))):
                warnings.warn(
                    f'The folder {folder_type}_{folder_key} does not exist. Creating it...')
        else:
            warnings.warn(
                f'Reference number of the folder {folder_type} not provided. '
                f'Overwriting the latest {folder_type} folder...')


class GeometricFunctions:

    @staticmethod
    def centroid(X, Y):
        """
            Computes the centroid coordinates of a non-self-intersecting closed polygon
            :param X: list of X coordinate of the vertices
            :param Y: list of y coordinate of the vertices
        """
        sum_A, sum_Cx, sum_Cy = 0, 0, 0
        for i in range(len(X)):
            index = i + 1 if i != len(X) - 1 else 0
            A = X[i] * Y[index] - X[index] * Y[i]
            sum_Cx += (X[i] + X[index]) * A
            sum_Cy += (Y[i] + Y[index]) * A
            sum_A += A
        factor = 1 / (3 * sum_A)
        return [factor * sum_Cx, factor * sum_Cy]

    @staticmethod
    def arc_center_from_3_points(a, b, c):
        """
            Computes the center coordinates of an arc passing through three points
            :param a: DataRoxieParser.Coord class object of one arc point
            :param b: DataRoxieParser.Coord class object of one arc point
            :param c: DataRoxieParser.Coord class object of one arc point
        """
        ab = [a.x - b.x, a.y - b.y]
        ac = [a.x - c.x, a.y - c.y]
        sac = [a.x * a.x - c.x * c.x, a.y * a.y - c.y * c.y]
        sba = [b.x * b.x - a.x * a.x, b.y * b.y - a.y * a.y]
        yy = (sac[0] * ab[0] + sac[1] * ab[0] + sba[0] * ac[0] + sba[1] * ac[0]) / \
             (2 * ((c.y - a.y) * ab[0] - (b.y - a.y) * ac[0]))
        xx = (sac[0] * ab[1] + sac[1] * ab[1] + sba[0] * ac[1] + sba[1] * ac[1]) / \
             (2 * ((c.x - a.x) * ab[1] - (b.x - a.x) * ac[1]))
        return [-xx, -yy]


class GmshUtils:

    def __init__(self, model_name=None, verbose=True):
        self.model_name = model_name
        self.verbose = verbose

    @staticmethod
    def initialize():
        if not gmsh.is_initialized():
            gmsh.initialize(sys.argv)
            num_threads = multiprocessing.cpu_count()
            gmsh.option.setNumber('General.NumThreads', num_threads)  # enable multithreading (this seems to be only for meshing)
            gmsh.option.setNumber('Mesh.MaxNumThreads1D', num_threads)
            gmsh.option.setNumber('Mesh.MaxNumThreads2D', num_threads)
            gmsh.option.setNumber('Mesh.MaxNumThreads3D', num_threads)

            gmsh.option.setNumber('Geometry.ToleranceBoolean', 0.0000001)
            gmsh.option.setNumber('General.Terminal', 1)
            # gmsh.model.add(self.model_name)

    def check_for_event(self):  # pragma: no cover
        action = gmsh.onelab.getString("ONELAB/Action")
        if len(action) and action[0] == "check":
            gmsh.onelab.setString("ONELAB/Action", [""])
            if self.verbose:
                print("-------------------check----------------")
            gmsh.fltk.update()
            gmsh.graphics.draw()
        if len(action) and action[0] == "compute":
            gmsh.onelab.setString("ONELAB/Action", [""])
            if self.verbose:
                print("-------------------compute----------------")
            gmsh.onelab.setChanged("Gmsh", 0)
            gmsh.onelab.setChanged("GetDP", 0)
            gmsh.fltk.update()
            gmsh.graphics.draw()
        return True

    def launch_interactive_GUI(self, close_after=-1):  # pragma: no cover
        gmsh.fltk.initialize()
        while gmsh.fltk.isAvailable() and self.check_for_event():
            gmsh.fltk.wait()
            if close_after >= 0:
                sleep(close_after)
                gmsh.fltk.finalize()
        gmsh.finalize()


class RoxieParsers:
    def __init__(self, conductor, block, xyCorner):
        self.conductor = conductor
        self.block = block
        self.xyCorner = xyCorner

    @staticmethod
    def parseMap2d(map2dFile: Path, headerLines: int = 1):
        """
            Generates array-stream of values of map2dFile
            :param map2dFile: path of map2dFile containing the content to parse
            :param headerLines: which index the header line is at - will start to read after that
        """
        # Open map2dfile
        fileContent = open(map2dFile, "r").read()
        # Split content of file in rows
        fileContentByRow = fileContent.split("\n")
        # Create array-matrix to fill in with the values of the file
        output_matrix = np.array([[None for x in range(10)] for y in range(headerLines + 1, len(fileContentByRow) - 1)],
                                 dtype=float)

        # Assign values to the matrix row by row
        for index, rowContent in enumerate(fileContentByRow):
            if index > headerLines and rowContent:  # without header
                row = rowContent.split()
                output_array = np.array([])  # create temp. array
                output_array = np.append(output_array, int(row[0]))  # strands to groups
                output_array = np.append(output_array, int(row[1]))  # strands to halfturn
                output_array = np.append(output_array, float(row[2]))  # idx
                output_array = np.append(output_array, float(row[3]) / 1e3)  # x_strands in [m]
                output_array = np.append(output_array, float(row[4]) / 1e3)  # y_strands in [m]
                output_array = np.append(output_array, float(row[5]))  # Bx
                output_array = np.append(output_array, float(row[6]))  # By
                output_array = np.append(output_array, float(row[7]) / 1e6)  # Area in [m^2]
                output_array = np.append(output_array, float(row[8]))  # I_strands
                output_array = np.append(output_array, float(row[9]))  # fill factor
                output_matrix[index - headerLines - 1] = output_array  # assign into matrix
        return output_matrix

    @staticmethod
    def parseCond2d(cond2dFile: Path):
        """
            Read input file and return list of ConductorPosition objects

            # input: fileName
            # output: conductorPositionsList

        """
        # conductorStartKeyword = "CONDUCTOR POSITION IN THE CROSS-SECTION"
        blockStartKeyword = "BLOCK POSITION IN THE CROSS-SECTION"

        fileContent = open(cond2dFile, "r").read()

        # separate rows
        fileContentByRow = fileContent.split("\n")

        # Find block definition
        for i in range(len(fileContentByRow)):
            if blockStartKeyword in fileContentByRow[i]:
                startOfBlockDefinitionIndex = i

        # separate part of the data with conductor position information
        conductorPositions = fileContentByRow[5:startOfBlockDefinitionIndex - 2]

        # drop every 5th row
        conductorPositionsFourVertices = list(conductorPositions)
        del conductorPositionsFourVertices[4::5]

        # arrange data in a list of lists
        outputConductorPositions = []
        for row in conductorPositionsFourVertices:
            rowSplitStr = row.split(',')
            rowSplitFloat = [float(elem) for elem in rowSplitStr]
            outputConductorPositions.append(rowSplitFloat)

        # arrange data from list to numpy.array
        outputConductorPositionsMatrix = np.array(outputConductorPositions)

        # input: outputConductorPositions
        # output: conductorPositionsList
        conductorPositionsList = []
        for i in range(0, len(outputConductorPositions), 4):
            out = outputConductorPositions[i]
            conductor = int(out[1])
            block = int(out[2])
            xyCorner = outputConductorPositionsMatrix[i:i + 4, 4:6]
            conductorPositionsList.append(RoxieParsers(conductor, block, xyCorner))

        return conductorPositionsList

def initialize_logger(verbose: bool = True, work_folder: str = None, time_stamp: str = None):
    logger = logging.getLogger()

    if verbose:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    for handler in logger.handlers:
        logger.handlers.remove(handler)
        handler.close()
    
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.INFO)
    stdout_handler.setFormatter(LoggingFormatter())
    logger.addHandler(stdout_handler)

    FilesAndFolders.prep_folder(work_folder)
    FilesAndFolders.prep_folder(os.path.join(work_folder, "logs"))
    file_handler = logging.FileHandler(os.path.join(work_folder, "logs", f"{time_stamp}.FiQuS.log"))
    file_handler.setLevel(logging.INFO)
    fileFormatter = logging.Formatter('%(asctime)s | %(levelname)s | %(message)s')
    file_handler.setFormatter(fileFormatter)
    logger.addHandler(file_handler)

    errorsAndWarnings_file_handler = logging.FileHandler(os.path.join(work_folder, "logs", f"ERRORS_WARNINGS_{time_stamp}.FiQuS.log"))
    errorsAndWarnings_file_handler.setLevel(logging.WARNING)
    errorsAndWarnings_file_handler.setFormatter(fileFormatter)
    logger.addHandler(errorsAndWarnings_file_handler)

    return logger



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    required = f.read().splitlines()

docs_require = [
    "markdown==3.3.7",
    "markdown-include==0.7.0",
    "mkdocs==1.3.1",
    "mkdocs-material==8.5.3",
    "mkdocstrings-python==0.7.1",
    "mkdocs-include-markdown-plugin==3.8.1",
    "mkdocs-git-revision-date-localized-plugin==1.1.0",
    "griffe==0.29.1",
]

tests_require = [
    "mypy==0.971",
    "pylint==2.15.3",
    "pytest==7.1.3",
    "pytest-cov==3.0.0",
    "coverage==6.4.4",
    "coverage-badge==1.1.0",
    "flake8==5.0.4",
]

build_require = [
    "setuptools==65.3.0",
]

all_requirements = required + docs_require + tests_require + build_require

setup(
    name="fiqus",
    version="2023.12.0",
    author="STEAM Team",
    author_email="steam-team@cern.ch",
    description="Source code for STEAM FiQuS tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.cern.ch/steam/fiqus",
    keywords={"STEAM", "FiQuS", "CERN"},
    install_requires=required,
    extras_require={
        "all": all_requirements,
        "docs": docs_require,
        "tests": tests_require,
        "build": build_require,
    },
    python_requires=">=3.8",
    package_data={
        "fiqus": [
            "CCT_template.pro",
            "Multipole_template.pro",
            "Pancake3D_template.pro",
            "materials.pro",
        ]
    },
    include_package_data=True,
    packages=find_packages(),
    classifiers=["Programming Language :: Python :: 3.8"],
)
